/*
 * Copyright (c) 2023-2024 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution  and  use  in  source  and binary  forms,  with  or  without
 * modification, are permitted provided that the following conditions are met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 * THE SOFTWARE  IS PROVIDED "AS IS"  AND THE AUTHOR  DISCLAIMS ALL WARRANTIES
 * WITH  REGARD   TO  THIS  SOFTWARE  INCLUDING  ALL   IMPLIED  WARRANTIES  OF
 * MERCHANTABILITY AND  FITNESS.  IN NO EVENT  SHALL THE AUTHOR  BE LIABLE FOR
 * ANY  SPECIAL, DIRECT,  INDIRECT, OR  CONSEQUENTIAL DAMAGES  OR  ANY DAMAGES
 * WHATSOEVER  RESULTING FROM  LOSS OF  USE, DATA  OR PROFITS,  WHETHER  IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR  OTHER TORTIOUS ACTION, ARISING OUT OF OR
 * IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *                                           Anthony Mallet on Wed Apr 12 2023
 */
#include "autoconf.h"

#include "tk3-paparazzi.h"

/* settings */
tk3setting(TK3SET_SRVMODE, 0, "0=off, 1=PWM, 2=DSHOT, 3=BI-DSHOT", 20);
tk3setting(TK3SET_SRV3D, 0, "0=off, 1=reversible", 21);
tk3setting(TK3SET_SRVFREQ, 150, "PWM=50-490Hz, DSHOT=100-900kHz", 22);
tk3setting(TK3SET_SRVIDLE, 6, "idle throttle (0%-20%)", 23);

/* DSHOT commands */
#define DSHOT_CMD_BEEP1			1
#define DSHOT_CMD_BEEP2			2
#define DSHOT_CMD_BEEP3			3
#define DSHOT_CMD_BEEP4			4
#define DSHOT_CMD_BEEP5			5
#define DSHOT_CMD_3D_MODE_OFF		9
#define DSHOT_CMD_3D_MODE_ON		10
#define DSHOT_CMD_SAVE_SETTINGS		12
#define DSHOT_EXTENDED_TELEMETRY_ENABLE	13

/* telemetry messages */
struct tlmmsg {
  enum {
    TK3TLM_NONE, TK3TLM_RPM, TK3TLM_TEMP, TK3TLM_VOLT, TK3TLM_AMP
  } kind;
  uint32_t value;
};

/* servo state */
static struct srvcfg {
  int hwinit;			/* hardware initialized */

  enum tk3srvmode mode;		/* control mode */
  uint32_t freq;		/* control freq (Hz PWM or kHz DSHOT) */
  uint32_t pidloop;		/* control period (µs) */
  int rev3d;			/* reversible thrust */
  int zero, idle;		/* zero & idle thrusts */
  uint32_t pmul;		/* eRPM period multiplier */

  struct tk3timer ctrl;		/* control timer */

  struct srvmotor {
    struct tk3timer cmd;	/* init/startup timer */

    unsigned enabled:1;		/* hardware detected */
    unsigned starting:1;
    unsigned spinning:1;
    unsigned servo:1;		/* velocity PID */
    unsigned feedback:1;	/* eRPM valid */

    union {
      int16_t throt;		/* current throttle */
      uint16_t raw;		/* raw DSHOT command */
    };
    float target;		/* desired eHz period (Hz) */
    uint32_t period;		/* current eRPM period (µs) */
    uint16_t current;		/* current draw (mA) */

    float pid;			/* PID output */
    float m0, m1;		/* PID measurements */
    float err0;			/* PID error history */
    float a0, a1, a0d, a1d;	/* PID parameters */
    float d0, fd0, a0fd, a1fd;	/* PID derivative filter */
  } motor[8];

  void (*arm)(void *);		/* mode-dependent routines */
  void (*tlm)(void);
  uint16_t (*map)(int16_t);
} srv;

static uint16_t	tk3srv_pwmmapthrottle(int16_t throttle);
static uint16_t	tk3srv_pwm3dmapthrottle(int16_t throttle);
static uint16_t	tk3srv_dshotmapthrottle(int16_t throttle);
static uint16_t	tk3srv_dshot3dmapthrottle(int16_t throttle);

static void	tk3srv_armpwm(void *data);
static void	tk3srv_armdshot(void *data);
static void	tk3srv_started(void *data);
static void	tk3srv_stopped(void *dummy);

static void	tk3srv_tlm(void);
static void	tk3srv_servo(struct srvmotor *motor);

static void	tk3srv_ctrlloop(void *data);


/* --- tk3srv_init --------------------------------------------------------- */

int
tk3srv_init()
{
  uint16_t idle;
  int i;

  /* reset state */
  srv = (struct srvcfg){0};
  srv.rev3d = tk3set_get(TK3SET_SRV3D, 0);
  srv.pmul = tk3set_get(TK3SET_POLES, 14) / 2;
  idle = tk3set_get(TK3SET_SRVIDLE, 5);
  if (idle > 20) {
    tk3msg_log(TK3CH_USB0, "Einvalid " TK3SET_SRVIDLE " setting");
    goto err;
  }
  srv.idle = (1<<15) * idle / 100;

  srv.mode = tk3set_get(TK3SET_SRVMODE, 0);
  srv.freq = tk3set_get(TK3SET_SRVFREQ, 600);

  switch(srv.mode) {
    case TK3SRV_OFF: return 0; /* no servo, done */

    case TK3SRV_PWM:
      /* PWM, frequency limited to below 500Hz */
      if (srv.freq < 50 || srv.freq > 490) {
        tk3msg_log(TK3CH_USB0, "Einvalid " TK3SET_SRVFREQ " setting");
        goto err;
      }
      srv.pidloop = 1000000/srv.freq;
      srv.arm = tk3srv_armpwm;
      srv.map =
        srv.rev3d ? tk3srv_pwm3dmapthrottle : tk3srv_pwmmapthrottle;
      srv.zero = srv.map(0);
      break;

    case TK3SRV_DSHOT: case TK3SRV_BIDSHOT:
      /* DSHOT modes, allow any freq within range */
      if (srv.freq < 100 || srv.freq > 900) {
        tk3msg_log(TK3CH_USB0, "Einvalid servo-freq settings");
        goto err;
      }
      srv.pidloop = 1000000/(srv.freq * 1000 / 100);
      srv.arm = tk3srv_armdshot;
      srv.map =
        srv.rev3d ? tk3srv_dshot3dmapthrottle : tk3srv_dshotmapthrottle;
      srv.zero = 0;
      if (srv.mode == TK3SRV_BIDSHOT) srv.tlm = tk3srv_tlm;
      break;

    default:
      tk3msg_log(TK3CH_USB0, "Einvalid servo-mode settings");
      goto err;
  }

  /* init default parameters */
  for(i = 0; i < 8; i++)
    tk3srv_setpid(i, 300., 1000., 0., 100.);

  return 0;
err:
  tk3fb_on(TK3FB_ESC_ERR);
  return 1;
}


/* --- tk3srv_fini --------------------------------------------------------- */

int
tk3srv_fini()
{
  int i;

  with_evlock() {
    for(i = 0; i < 8; i++) tk3clk_deltimer(&srv.motor[i].cmd);
    tk3clk_deltimer(&srv.ctrl);

    if (srv.hwinit) tk3srv_hwfini();
    srv.hwinit = 0;
  }

  return 0;
}


/* --- tk3srv_detect ------------------------------------------------------- */

void
tk3srv_detect()
{
  unsigned int d;
  int i;

  with_evlock() {
    if (srv.hwinit) return;

    /* detect ESC presence */
    for(i = 0; i < 8; i++) {
      d = srv.motor[i].enabled;
      srv.motor[i].enabled = tk3srv_hwdetect(i);
      if (d != srv.motor[i].enabled) {
        if (srv.motor[i].enabled)
          tk3msg_log(TK3CH_USB0, "NESC #%c detected", '1' + i);
        else
          tk3msg_log(TK3CH_USB0, "NESC #%c lost", '1' + i);
      }
    }
  }
}


/* --- tk3srv_dshotencode -------------------------------------------------- */

static uint16_t
tk3srv_dshotencode(uint16_t throt, bool tlm, uint8_t rev)
{
  uint16_t msg, crc;

  /* pack throt and telemetry bit */
  msg = ((throt & 0x7ff) << 1) | !!tlm;

  /* telemetry bit needs to be set for special commands */
  msg |= throt > 0 && throt < 48;

  /* xor nibbles */
  crc = msg ^ (msg >> 4) ^ (msg >> 8);

  /* pack frame, reverting CRC for bi-dshot */
  return (msg << 4) | ((crc ^ rev) & 0xf);
}

uint16_t
tk3srv_dshotmsg(uint16_t throt, bool tlm)
{
  return tk3srv_dshotencode(throt, tlm, 0 /* normal checksum */);
}

uint16_t
tk3srv_bidshotmsg(uint16_t throt, bool tlm)
{
  return tk3srv_dshotencode(throt, tlm, 0xf /* inverted cksum */);
}


/* --- tk3srv_dshotdecode -------------------------------------------------- */

static struct tlmmsg
tk3srv_dshotdecode(uint32_t tlm)
{
  uint16_t crc;

  /* get GCR */
  tlm = tlm ^ (tlm >> 1);

  /* decode GCR */
  static const uint8_t gcr[32] = {
    [0 ... 0x1f] = -1, /* invalid by default */

    [0x19] = 0x0, [0x1b] = 0x1, [0x12] = 0x2, [0x13] = 0x3,
    [0x1d] = 0x4, [0x15] = 0x5, [0x16] = 0x6, [0x17] = 0x7,
    [0x1a] = 0x8, [0x09] = 0x9, [0x0a] = 0xa, [0x0b] = 0xb,
    [0x1e] = 0xc, [0x0d] = 0xd, [0x0e] = 0xe, [0x0f] = 0xf
  };

  uint16_t erpm =
    gcr[tlm & 0x1f] |
    (gcr[(tlm >> 5) & 0x1f] << 4) |
    (gcr[(tlm >> 10) & 0x1f] << 8) |
    (gcr[(tlm >> 15) & 0x1f] << 12);

  /* check CRC */
  crc = erpm ^ (erpm >> 8);
  crc ^= (crc >> 4);
  if (~crc & 0xf) return (struct tlmmsg){ .kind = TK3TLM_NONE };

  erpm >>= 4;

  /* decode */
  switch(erpm & 0xf00) {
    case 0x200:
      return (struct tlmmsg){ .kind = TK3TLM_TEMP };

    case 0x400:
      return (struct tlmmsg){ .kind = TK3TLM_VOLT };

    case 0x600:
      return (struct tlmmsg){ .kind = TK3TLM_AMP, .value = erpm & 0xff };

    case 0x000:
    case 0x800:
    case 0xa00:
    case 0xc00:
    case 0xe00:
      return (struct tlmmsg){ .kind = TK3TLM_NONE };

    default:
      if ((erpm & 0xfff) == 0xfff) /* saturated */
        return (struct tlmmsg){ .kind = TK3TLM_NONE };

      return (struct tlmmsg){
        .kind = TK3TLM_RPM,
        .value = (erpm & 0x1ff) << ((erpm & 0xe00) >> 9)
      };
  }
}


/* --- tk3srv_mapthrottle -------------------------------------------------- */

static uint16_t
tk3srv_pwmmapthrottle(int16_t throttle)
{
  return throttle < 0 ? 0 : throttle;
}

static uint16_t
tk3srv_pwm3dmapthrottle(int16_t throttle)
{
  return (1 << 14) + (throttle >> 1);
}

static uint16_t
tk3srv_dshotmapthrottle(int16_t throttle)
{
  return 48 +
    (((throttle < 0 ? 0 : throttle) * 2000) >> 15);
}

static uint16_t
tk3srv_dshot3dmapthrottle(int16_t throttle)
{
  return 48 +
    (throttle < 0 ? 1000 : 0) +
    (((throttle < 0 ? -throttle : throttle) * 1000) >> 15);
}


/* --- tk3srv_tlm ---------------------------------------------------------- */

static void
tk3srv_tlm()
{
  uint32_t tlm[8];
  struct tlmmsg m;
  int i;

  tk3srv_read(tlm);

  with_evlock() {
    for(i = 0; i < 8; i++) {
      srv.motor[i].feedback = 0;

      m = tk3srv_dshotdecode(tlm[i]);
      switch (m.kind) {
        default: break;

        case TK3TLM_AMP:
          srv.motor[i].current = m.value * 1000;
          break;

        case TK3TLM_RPM:
          srv.motor[i].period = m.value;
          srv.motor[i].enabled = srv.motor[i].feedback = 1;
          break;
      }
    }
  }
}


/* --- tk3srv_arm ---------------------------------------------------------- */

static void
tk3srv_armpwm(void *data)
{
  struct srvmotor *motor = data;

  coroutinex(motor - srv.motor, 8) {
    /* zero throttle a bit */
    motor->raw = srv.zero;
    tk3clk_settimer(
      &motor->cmd, 0, 5000000 /* 5s */, TK3EV_IDLE, tk3srv_armpwm, data);
    yield();

    /* armed */
    tk3srv_started(data);
  }
}

static void
tk3srv_armdshot(void *data)
{
  struct srvmotor *motor = data;

  coroutinex(motor - srv.motor, 8) {
    /* zero - wait for detection */
    motor->raw = 0;
    tk3clk_settimer(
      &motor->cmd, 0, 3000000 /* 3s */, TK3EV_IDLE, tk3srv_armdshot, data);
    yield();

    /* set extended telemetry mode */
    motor->raw = DSHOT_EXTENDED_TELEMETRY_ENABLE;
    tk3clk_settimer(
      &motor->cmd, 0, 12*srv.pidloop, TK3EV_IDLE, tk3srv_armdshot, data);
    yield();
    motor->raw = 0;
    tk3clk_settimer(
      &motor->cmd, 0, 35000 /* 35ms */, TK3EV_IDLE, tk3srv_armdshot, data);
    yield();

    /* set 3D mode */
    motor->raw = srv.rev3d ? DSHOT_CMD_3D_MODE_ON : DSHOT_CMD_3D_MODE_OFF;
    tk3clk_settimer(
      &motor->cmd, 0, 12*srv.pidloop, TK3EV_IDLE, tk3srv_armdshot, data);
    yield();

    /* save settings (required for some firmware) */
    motor->raw = DSHOT_CMD_SAVE_SETTINGS;
    tk3clk_settimer(
      &motor->cmd, 0, 12*srv.pidloop, TK3EV_IDLE, tk3srv_armdshot, data);
    yield();

    /* stop motors */
    motor->raw = 0;
    tk3clk_settimer(
      &motor->cmd, 0, 250000 /* 250ms */, TK3EV_IDLE, tk3srv_armdshot, data);
    yield();

    /* armed */
    tk3srv_started(data);
  }
}


/* --- tk3srv_start -------------------------------------------------------- */

int
tk3srv_start(int index)
{
  int i, l, h;

  l = index < 0 ? 0 : index;
  h = index < 0 ? 7 : index;

  /* detect ESCs */
  tk3srv_detect();

  with_evlock() {
    if (!srv.hwinit) {
      /* init lines */
      if (tk3srv_hwinit(srv.mode, srv.freq)) return 1;

      /* enable control signal */
      tk3clk_settimer(
        &srv.ctrl, srv.pidloop, 0, TK3EV_NOW, tk3srv_ctrlloop, NULL);
      srv.hwinit = 1;
    }

    for(i = l; i <= h; i++) {
      if (!srv.motor[i].enabled) continue;
      if (srv.motor[i].spinning) continue;

      srv.motor[i].starting = 1;
      srv.arm(&srv.motor[i]);
    }
  }

  return 0;
}

static void
tk3srv_started(void *data)
{
  struct srvmotor *motor = data;

  coroutinex(motor - srv.motor, 8) {
    /* wait a bit to reach idle velocity */
    motor->raw = srv.map(srv.idle);
    tk3clk_settimer(
      &motor->cmd, 0, 500000 /* 500ms */, TK3EV_IDLE, tk3srv_started, data);
    yield();

    /* started */
    with_evlock() {
      motor->spinning = 1;
      motor->starting = 0;
      motor->throt = srv.idle;

      /* reset PID state */
      motor->servo = 0;
      motor->target = 0.;
      motor->pid = srv.idle;
      motor->m0 = motor->m1 = 0.;
      motor->err0 = 0.;
      motor->d0 = motor->fd0 = 0.;
    }
  }
}


/* --- tk3srv_stop --------------------------------------------------------- */

int
tk3srv_stop(int index)
{
  int i, l, h;

  l = index < 0 ? 0 : index;
  h = index < 0 ? 7 : index;

  with_evlock() {
    for(i = l; i <= h; i++) {
      if (!srv.motor[i].enabled) continue;

      srv.motor[i].spinning = 0;
      srv.motor[i].starting = 0;
      srv.motor[i].raw = srv.zero;
      srv.motor[i].period = 0;
    }

    /* wait a bit to reach 0 velocity before disabling hw */
    tk3clk_settimer(&srv.motor[l].cmd, 0, 100000 /* 100ms */,
                    TK3EV_IDLE, tk3srv_stopped, NULL);
  }

  return 0;
}

static void
tk3srv_stopped(void *dummy)
{
  int i;

  with_evlock() {
    for(i = 0; i < 8; i++) {
      if (!srv.motor[i].enabled) continue;
      if (srv.motor[i].starting || srv.motor[i].spinning)
        return;
    }
  }

  tk3srv_fini();
}


/* --- tk3srv_setpid ------------------------------------------------------- */

void
tk3srv_setpid(int index, float Kp, float Ki, float Kd, float dfc)
{
  if (index < 0 || index >= sizeof(srv.motor)/sizeof(*srv.motor))
    return;

  double dt = srv.pidloop / 1e6;	/* control loop period */
  double tau = 6.2832 * dfc * dt;	/* LP filter ωc * dt */

  with_evlock() {
    srv.motor[index].a0 = Kp + Ki * dt;
    srv.motor[index].a1 = -Kp;

    srv.motor[index].a0d = Kd / dt;
    srv.motor[index].a1d = - 2 * Kd / dt;

    srv.motor[index].a0fd = tau / (2 + tau);	/* PID derivative filter */
    srv.motor[index].a1fd = (2 - tau) / (2 + tau);
  }
}


/* --- tk3srv_beep --------------------------------------------------------- */

void
tk3srv_beep(void *data)
{
  int i;

  /* this is only for DSHOT modes */
  switch (srv.mode) {
    default: return;

    case TK3SRV_DSHOT:
    case TK3SRV_BIDSHOT: break;
  }

  with_evlock() {
    for(i = 0; i < 8; i++) {
      if (!srv.motor[i].enabled) continue;
      if (srv.motor[i].starting || srv.motor[i].spinning) continue;

      if (!data)
        srv.motor[i].raw = srv.zero;
      else {
        srv.motor[i].raw = DSHOT_CMD_BEEP1;
        tk3clk_settimer(&srv.motor[i].cmd, 0,
                        12 * srv.pidloop, TK3EV_IDLE, tk3srv_beep, NULL);
      }
    }
  }
}


/* --- tk3srv_throttle ----------------------------------------------------- */

void
tk3srv_throttle(const int16_t throt[8])
{
  int i;

  with_evlock() {
    if (!tk3rc_killswitch) /* normal case */
      for(i = 0; i < 8; i++) {
        if (!srv.motor[i].spinning) continue;

        srv.motor[i].servo = 0;
        srv.motor[i].throt = throt[i] << 5;
      }

    else /* kill switch active, stop motors */
      for(i = 0; i < 8; i++) {
        if (!srv.motor[i].spinning) continue;

        srv.motor[i].spinning = 0;
        srv.motor[i].raw = srv.zero;
      }
  }
}


/* --- tk3srv_velocity ----------------------------------------------------- */

void
tk3srv_velocity(const int16_t hperiod[8] /* half period */)
{
  float k = (float)1e6 * srv.pmul / 2;
  int i;

  with_evlock() {
    if (!tk3rc_killswitch) /* normal case */
      for(i = 0; i < 8; i++) {
        if (!srv.motor[i].spinning) continue;
        srv.motor[i].servo = 1;
        if (!srv.rev3d && hperiod[i] < 0)
          srv.motor[i].target = k / 0x7fff;
        else {
          if (__builtin_signbit(srv.motor[i].target) != (hperiod[i] >> 15)) {
            srv.motor[i].pid = srv.idle;
            srv.motor[i].m0 = srv.motor[i].m1 = 0.;
            srv.motor[i].err0 = 0.;
            srv.motor[i].d0 = srv.motor[i].fd0 = 0.;
          }
          srv.motor[i].target = k / hperiod[i];
        }
      }

    else /* kill switch active, stop motors */
      for(i = 0; i < 8; i++) {
        if (!srv.motor[i].spinning) continue;

        srv.motor[i].spinning = 0;
        srv.motor[i].raw = srv.zero;
      }
  }
}


/* --- tk3srv_motdata ------------------------------------------------------ */

void
tk3srv_motdata(void *arg)
{
  static uint8_t seq;
  uint16_t period, pwm, current;
  uint8_t state;
  uint32_t v;
  int i;

  for(i = 0; i < 8; i++) {
    if (!srv.motor[i].enabled) continue;

    with_evlock() {
      state =
        (srv.motor[i].spinning << 5) |
        (srv.motor[i].starting << 4) |
        ((i + 1) & 0xf);

      v = srv.motor[i].period * srv.pmul;
      if (v >= 0xffff)
        period = (srv.motor[i].spinning ? 0xffff : 0)/2;
      else
        period = v/2;
      if (srv.motor[i].throt < 0) period = -period;

      pwm = srv.motor[i].throt >> 5;

      current = srv.motor[i].current;
    }

    tk3msg_log(TK3CH_USB0, "M%1%1%2%2%2", seq, state, period, pwm, current);
  }
  seq++;
}


/* --- tk3srv_servo -------------------------------------------------------- */

static void
tk3srv_servo(struct srvmotor *motor)
{
  float m2;
  float err1, d1, fd1;

  /* Discrete PID implementation, see https://en.wikipedia.org/wiki/
   * Proportional%E2%80%93integral%E2%80%93derivative_controller#Pseudocode
   */
  m2 = motor->m1;
  motor->m1 = motor->m0;
  motor->m0 = (float)1e6/motor->period;

  err1 = motor->err0;
  motor->err0 = __builtin_fabsf(motor->target) - motor->m0;

  /* kP, kI */
  motor->pid += motor->a0 * motor->err0 + motor->a1 * err1;

  /* filtered kD */
  if (motor->a0d) {
    d1 = motor->d0;
    fd1 = motor->fd0;

    motor->d0 =
      motor->a0d * motor->m0 + motor->a1d * motor->m1 + motor->a0d * m2;
    motor->fd0 = motor->a0fd * (motor->d0 + d1) + motor->a1fd * fd1;
    motor->pid += motor->fd0;
  }

  if (motor->pid > 32767.) motor->pid = 32767.;
  else if (motor->pid < -32768.) motor->pid = -32768.;

  motor->throt = motor->pid < 1. ? 1 : motor->pid;
  if (__builtin_signbit(motor->target))
      motor->throt = -motor->throt;
}


/* --- tk3srv_ctrlloop ----------------------------------------------------- */

static void
tk3srv_ctrlloop(void *data)
{
  uint16_t target[8];
  int i;

  /* update telemetry if enabled */
  if (srv.tlm) srv.tlm();

  /* compute target throttle */
  with_evlock() {
    for(i = 0; i < 8; i++) {
      if (srv.motor[i].spinning) {
        if (srv.motor[i].servo && srv.motor[i].feedback)
          tk3srv_servo(&srv.motor[i]);

        target[i] = srv.map(srv.motor[i].throt);
      } else
        target[i] = srv.motor[i].raw;
    }
  }

  tk3srv_send(target);
}
